"""Dynamical climate downscaling using deep convolutional neural networks."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# builtins
import sys
import time
import logging
from datetime import timedelta
from logging.config import dictConfig

# externals
import xarray as xr

# locals
from pysegcnn.core.trainer import LogConfig
from pysegcnn.core.models import Network
from pysegcnn.core.logging import log_conf
from pysegcnn.core.utils import search_files
from climax.core.dataset import ERA5Dataset
from climax.core.predict import predict_ERA5
from climax.core.utils import split_date_range
from climax.main.config import (ERA5_PREDICTORS, ERA5_PLEVELS, PREDICTAND, NET,
                                VALID_PERIOD, BATCH_SIZE, NORM, DOY, NYEARS,
                                DEM, DEM_FEATURES, LOSS)
from climax.main.io import ERA5_PATH, DEM_PATH, MODEL_PATH, TARGET_PATH

# module level logger
LOGGER = logging.getLogger(__name__)


if __name__ == '__main__':

    # initialize timing
    start_time = time.monotonic()

    # initialize network filename
    state_file = ERA5Dataset.state_file(
        NET, PREDICTAND, ERA5_PREDICTORS, ERA5_PLEVELS, dem=DEM,
        dem_features=DEM_FEATURES, doy=DOY, loss=LOSS)

    # path to model state
    state_file = MODEL_PATH.joinpath(PREDICTAND, state_file)

    # initialize logging
    log_file = MODEL_PATH.joinpath(PREDICTAND,
                                   state_file.name.replace('.pt', '_log.txt'))
    dictConfig(log_conf(log_file))

    # predict reference period
    LogConfig.init_log('Predicting reference period: {}'.format(
        ' - '.join([str(VALID_PERIOD[0]), str(VALID_PERIOD[-1])])))

    # initialize ERA5 predictor dataset
    LogConfig.init_log('Initializing ERA5 predictors.')
    Era5 = ERA5Dataset(ERA5_PATH.joinpath('ERA5'), ERA5_PREDICTORS,
                       plevels=ERA5_PLEVELS)
    Era5_ds = Era5.merge(chunks=-1)

    # whether to use digital elevation model
    if DEM:
        # digital elevation model: Copernicus EU-Dem v1.1
        dem = search_files(DEM_PATH, '^eu_dem_v11_stt.nc$').pop()

        # read elevation and compute slope and aspect
        dem = ERA5Dataset.dem_features(
            dem, {'y': Era5_ds.y, 'x': Era5_ds.x},
            add_coord={'time': Era5_ds.time})

        # check whether to use slope and aspect
        if not DEM_FEATURES:
            dem = dem.drop_vars(['slope', 'aspect'])

        # add dem to set of predictor variables
        Era5_ds = xr.merge([Era5_ds, dem])

    # load pretrained model
    if state_file.exists():
        # load pretrained network
        net, _ = Network.load_pretrained_model(state_file, NET)
    else:
        # initialize OBS predictand dataset
        LOGGER.info('{} does not exist.'.format(state_file))
        sys.exit()

    # subset to reference period and predict in NYEAR intervals
    trg_ds = []
    for dates in split_date_range(VALID_PERIOD[0], VALID_PERIOD[-1],
                                  years=NYEARS):
        LogConfig.init_log('Predicting period: {}'.format(
            ' - '.join([str(dates[0]), str(dates[-1])])))
        ref_ds = Era5_ds.sel(time=dates)
        trg_ds.append(predict_ERA5(net, ref_ds, PREDICTAND, LOSS,
                                   normalize=NORM, batch_size=BATCH_SIZE,
                                   doy=DOY))

    # merge predictions for entire validation period
    LOGGER.info('Merging reference periods ...')
    trg_ds = xr.concat(trg_ds, dim='time')

    # save model predictions as NetCDF file
    target = TARGET_PATH.joinpath(PREDICTAND, net.state_file.name.replace(
        net.state_file.suffix, '.nc'))
    if not target.parent.exists():
        target.parent.mkdir(parents=True, exist_ok=True)
    LOGGER.info('Saving network predictions: {}.'.format(target))
    trg_ds.to_netcdf(target, engine='h5netcdf')

    # log execution time of script
    LogConfig.init_log('Execution time of script {}: {}'
                       .format(__file__, timedelta(seconds=time.monotonic() -
                                                   start_time)))
